import ipywidgets as widgets
from IPython.display import display
from penelope.corpus.vectorized_corpus import VectorizedCorpus

from notebooks.word_trends.displayers import display_bar, display_grid, display_line, display_table


def display_gui(state):

    output_widget = widgets.Output(layout=widgets.Layout(width='600px', height='200px'))
    words_widget = widgets.Textarea(
        description="", rows=4, value="och eller hur", layout=widgets.Layout(width='600px', height='200px')
    )
    tab_widget = widgets.Tab()

    data_handlers = [display_table, display_line, display_bar, display_grid]
    tab_widget.children = [widgets.Output() for _ in data_handlers]
    _ = [tab_widget.set_title(i, x.NAME) for i, x in enumerate(data_handlers)]

    z_corpus: VectorizedCorpus = None
    x_corpus: VectorizedCorpus = None

    def update_plot(*_):

        nonlocal z_corpus, x_corpus

        if state.corpus is None:

            with output_widget:
                print("Please load a corpus!")

            return

        if z_corpus is None or z_corpus is not state.corpus:

            with output_widget:
                print("Corpus changed...")

            z_corpus = state.corpus
            x_corpus = z_corpus.todense()

            tab_widget.children[1].clear_output()
            with tab_widget.children[1]:
                display_line.setup(state, x_ticks=[x for x in x_corpus.xs_years()], plot_width=1000, plot_height=500)

        tokens = '\n'.join(words_widget.value.split()).split()
        tab_index = tab_widget.selected_index
        data_handler = data_handlers[tab_index]
        indices = [x_corpus.token2id[token] for token in tokens if token in x_corpus.token2id]

        with output_widget:

            if len(indices) == 0:
                print("Nothing to show!")
                return

            missing_tokens = [token for token in tokens if token not in x_corpus.token2id]
            if len(missing_tokens) > 0:
                print("Not in corpus subset: {}".format(' '.join(missing_tokens)))

        if data_handler.NAME != "Line":
            tab_widget.children[tab_index].clear_output()

        with tab_widget.children[tab_index]:

            data = data_handler.compile(x_corpus, indices)
            state.data = data
            data_handler.plot(data, container=state)

    words_widget.observe(update_plot, names='value')
    tab_widget.observe(update_plot, 'selected_index')

    display(widgets.VBox([widgets.HBox([words_widget, output_widget]), tab_widget]))

    update_plot()



from dataclasses import dataclass
from typing import Dict

import ipywidgets
import pandas as pd
import penelope.common.goodness_of_fit as gof
import penelope.notebook.utility as notebook_utility
import penelope.notebook.word_trend_plot_gui as word_trend_plot_gui
from penelope.co_occurrence.concept_co_occurrence import to_vectorized_corpus
from penelope.corpus import VectorizedCorpus
from penelope.notebook.ipyaggrid_utility import display_grid as display_as_grid
from tqdm import tqdm


@dataclass
class State:
    concept_co_occurrences: pd.DataFrame = None
    concept_co_occurrences_metadata: Dict = None
    corpus: VectorizedCorpus = None
    corpus_tag: str = None
    metadata: Dict = None
    goodness_of_fit: pd.DataFrame = None
    most_deviating_overview: pd.DataFrame = None
    most_deviating: pd.DataFrame = None


state = State()


def loaded_callback(
    output: ipywidgets.Output,
    *,
    concept_co_occurrences: pd.DataFrame = None,
    metadata: Dict = None,
    corpus: VectorizedCorpus = None,
    corpus_tag: str = None,
):
    global state

    pbar = tqdm(total=5)
    output.clear_output()
    with output:

        pbar.update(1)

        state.concept_co_occurrences = concept_co_occurrences
        state.metadata = metadata

        if corpus is None:

            if state.concept_co_occurrences is None:
                raise ValueError("Both corpus and concept_co_occurrences cannot be None")

            pbar.set_description("Vectorizing co-occurrences...")
            state.corpus = to_vectorized_corpus(
                co_occurrences=state.concept_co_occurrences, value_column='value_n_t'
            ).group_by_year()

            state.corpus_tag = corpus_tag
        pbar.update(1)

        pbar.set_description("Computing goodness of fit...")
        state.goodness_of_fit = gof.compute_goddness_of_fits_to_uniform(state.corpus, None, verbose=False)
        pbar.update(1)
        pbar.set_description("Most deviating...")
        state.most_deviating_overview = gof.compile_most_deviating_words(state.goodness_of_fit, n_count=10000)
        pbar.update(1)
        state.most_deviating = gof.get_most_deviating_words(
            state.goodness_of_fit, 'l2_norm', n_count=5000, ascending=False, abs_value=True
        )
        pbar.update(1)
        pbar.set_description("Setting upp GUI...")

        wtf = word_trend_plot_gui.display_gui(state.corpus, tokens=state.most_deviating, display_widgets=False)

        tab_gof = (
            notebook_utility.OutputsTabExt(["GoF", "GoF (abs)", "Plots", "Slopes"])
            .display_fx_result(0, display_as_grid, state.goodness_of_fit)
            .display_fx_result(
                1, display_as_grid, state.most_deviating_overview[['l2_norm_token', 'l2_norm', 'abs_l2_norm']]
            )
            .display_fx_result(2, gof.plot_metrics, state.goodness_of_fit, plot=False, lazy=True)
            .display_fx_result(
                3, gof.plot_slopes, state.corpus, state.most_deviating, "l2_norm", 600, 600, plot=False, lazy=True
            )
        )

        _ = (
            notebook_utility.OutputsTabExt(["Data", "Explore", "Options", "GoF"])
            .display()
            .display_fx_result(0, display_as_grid, concept_co_occurrences)
            .display_content(1, what=wtf, clear=True)
            .display_as_yaml(2, state.metadata, clear=True, width='800px', height='600px')
            .display_content(3, tab_gof, clear=True)
        )
        pbar.update(2)

        pbar.close()



import itertools

import bokeh


def take(n, iterable):
    "Return first n items of the iterable as a list"
    return list(itertools.islice(iterable, n))


def compile_multiline_data(x_corpus, indices, smoothers=None):

    xs = x_corpus.xs_years()

    if len(smoothers or []) > 0:
        xs_data = []
        ys_data = []
        for j in indices:
            xs_j = xs
            ys_j = x_corpus.bag_term_matrix[:, j]
            for smoother in smoothers:
                xs_j, ys_j = smoother(xs_j, ys_j)
            xs_data.append(xs_j)
            ys_data.append(ys_j)
    else:
        xs_data = [xs.tolist()] * len(indices)
        ys_data = [x_corpus.bag_term_matrix[:, token_id].tolist() for token_id in indices]

    data = {
        'xs': xs_data,
        'ys': ys_data,
        'label': [x_corpus.id2token[token_id].upper() for token_id in indices],
        'color': take(len(indices), itertools.cycle(bokeh.palettes.Category10[10])),
    }
    return data


def compile_year_token_vector_data(x_corpus, indices, *args):  # pylint: disable=unused-argument

    xs = x_corpus.xs_years()
    data = {x_corpus.id2token[token_id]: x_corpus.bag_term_matrix[:, token_id] for token_id in indices}
    data['year'] = xs

    return data



import pandas as pd
from ipyaggrid import Grid
from IPython.display import display

from . import data_compilers

NAME = "Grid"

compile = data_compilers.compile_year_token_vector_data  # pylint: disable=redefined-builtin


def setup(container, **kwargs):  # pylint: disable=unused-argument
    pass


def default_column_defs(df):
    column_defs = [
        {
            'headerName': column.title(),
            'field': column,
            # 'rowGroup':False,
            # 'hide':False,
            'cellRenderer': "function(params) { return params.value.toFixed(6); }" if column != 'year' else None,
            # 'type': 'numericColumn'
        }
        for column in df.columns
    ]
    return column_defs


def plot(data, **kwargs):  # pylint: disable=unused-argument

    df = pd.DataFrame(data=data).set_index('year')
    column_defs = default_column_defs(df)
    grid_options = {
        'columnDefs': column_defs,
        'enableSorting': True,
        'enableFilter': True,
        'enableColResize': True,
        'enableRangeSelection': False,
    }

    g = Grid(
        grid_data=df,
        grid_options=grid_options,
        quick_filter=False,
        show_toggle_edit=False,
        export_mode="buttons",
        export_csv=True,
        export_excel=False,
        theme='ag-theme-balham',
        show_toggle_delete=False,
        columns_fit='auto',
        index=True,
        keep_multiindex=False,
        menu={'buttons': [{'name': 'Export Grid', 'hide': True}]},
    )

    display(g)



from .ipyaggrid_utility import display_grid
from .model_container import TopicModelContainer
from .topic_utility import filter_by_key_value, filter_document_topic_weights
from .utils import setup_pandas, to_text



import itertools
import math

import bokeh

from . import data_compilers

NAME = "Line"

compile = data_compilers.compile_multiline_data  # pylint: disable=redefined-builtin


def setup(container, **kwargs):

    x_ticks = kwargs.get('x_ticks', None)
    plot_width = kwargs.get('plot_width', 1000)
    plot_height = kwargs.get('plot_height', 800)

    data = {'xs': [[0]], 'ys': [[0]], 'label': [""], 'color': ['red']}  # , 'token_id': [ 0 ] }

    data_source = bokeh.models.ColumnDataSource(data)

    p = bokeh.plotting.figure(plot_width=plot_width, plot_height=plot_height)
    p.y_range.start = 0
    p.yaxis.axis_label = 'Frequency'
    p.toolbar.autohide = True

    if x_ticks is not None:
        p.xaxis.ticker = x_ticks

    p.xaxis.major_label_orientation = math.pi / 4
    p.xgrid.grid_line_color = None
    p.ygrid.grid_line_color = None

    _ = p.multi_line(xs='xs', ys='ys', legend_field='label', line_color='color', source=data_source)

    p.legend.location = "top_left"
    p.legend.click_policy = "hide"
    p.legend.background_fill_alpha = 0.0

    container.figure = p
    container.handle = bokeh.plotting.show(p, notebook_handle=True)
    container.data_source = data_source


def plot(data, **_):

    years = [str(y) for y in data['year']]

    data['year'] = years

    tokens = [w for w in data.keys() if w != 'year']

    source = bokeh.models.ColumnDataSource(data=data)

    max_value = max([max(data[key]) for key in data if key != 'year']) + 0.005

    p = bokeh.plotting.figure(
        x_range=years, y_range=(0, max_value), plot_height=400, plot_width=1000, title="Word frequecy by year"
    )

    colors = itertools.islice(itertools.cycle(bokeh.palettes.d3['Category20b'][20]), len(tokens))

    offset = -0.25
    v = []
    for token in tokens:
        w = p.vbar(
            x=bokeh.transform.dodge('year', offset, range=p.x_range),
            top=token,
            width=0.2,
            source=source,
            color=next(colors),
        )  # , legend_label=token)
        offset += 0.25
        v.append(w)

    p.x_range.range_padding = 0.04
    p.xaxis.major_label_orientation = math.pi / 4
    p.xgrid.grid_line_color = None
    p.ygrid.grid_line_color = None
    p.legend.location = "top_right"
    p.legend.orientation = "vertical"

    legend = bokeh.models.Legend(items=[(x, [v[i]]) for i, x in enumerate(tokens)])

    p.add_layout(legend, 'left')

    bokeh.io.show(p)



from typing import Set

from penelope.co_occurrence import partitioned_corpus_concept_co_occurrence, to_vectorized_corpus
from penelope.co_occurrence.concept_co_occurrence import ConceptContextOpts
from penelope.corpus import AnnotationOpts, SparvTokenizedCsvCorpus


def generate_test_vectorized_corpus(
    filename: str,
    concepts: Set[str],
    *,
    context_width: int = 2,
    annotation_opts: AnnotationOpts = None,
    n_count_threshold: int = None,
    partition_keys: str = "year",
    no_concept: bool = False,
):
    annotation_opts = annotation_opts or AnnotationOpts()
    corpus = SparvTokenizedCsvCorpus(
        filename,
        tokenizer_opts=dict(
            filename_fields={"year": r"prot\_(\d{4}).*"},
        ),
        annotation_opts=annotation_opts,
    )

    if not any([concept in corpus.token2id for concept in concepts]):
        print("concept not found in corpus")
        return None

    coo_df = partitioned_corpus_concept_co_occurrence(
        corpus,
        concept_opts=ConceptContextOpts(concepts=concepts, no_concept=no_concept, n_context_width=context_width),
        n_count_threshold=n_count_threshold,
        partition_keys=partition_keys,
    )

    # store_co_occurrences('/data/riksdagens-protokoll.1920-2019.test', coo_df)
    _v_corpus = to_vectorized_corpus(coo_df, value_column="value_n_t")

    _v_corpus.dump(
        tag=f"coo_vectors_{'_'.join(concepts)}_{''.join(annotation_opts.get_pos_includes())}_{context_width}",
        folder="/data/westac/",
    )

    return _v_corpus


v_corpus_riksdagens_protokoll = generate_test_vectorized_corpus(
    "/data/westac/riksdagens-protokoll.1920-2019.test.zip",
    concepts={"arbetare"},
    context_width=2,
    annotation_opts=AnnotationOpts(lemmatize=True),
)

# %%



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.6.0
#   kernelspec:
#     display_name: 'Python 3.7.5 64-bit (''welfare_state_analytics'': pipenv)'
#     language: python
#     name: python37564bitwelfarestateanalyticspipenvb8730e518d45450a918e95a98e055dbf
# ---

# %% [markdown]
# ## Word Trend Analysis
#
# ### Load previously vectorized corpus
#
# Use the `vectorize_protocol` script to create a new corpus with different settings.

# %% tags=[] vscode={"end_execution_time": "2020-08-31T18:34:55.995Z", "start_execution_time": "2020-08-31T18:34:55.854Z"}
# %load_ext autoreload
# %autoreload 2


# %%

import importlib
import os
from dataclasses import dataclass
from typing import Any

import pandas as pd
import penelope.common.goodness_of_fit as gof
import penelope.notebook.load_vectorized_corpus_gui as load_corpus_gui
import penelope.notebook.utility as notebook_utility
import penelope.notebook.vectorize_corpus_gui as vectorize_corpus_gui
import penelope.notebook.word_trend_plot_gui as word_trend_plot_gui
from bokeh.plotting import output_notebook
from penelope.corpus import VectorizedCorpus
from penelope.notebook.ipyaggrid_utility import display_grid

import __paths__  # pylint: disable=unused-import
import notebooks.word_trends.word_trends_gui as xxx_word_trends_gui


@dataclass
class State:
    corpus_folder: str = '/data/westac'
    corpus: VectorizedCorpus = None
    data: Any = None
    df_gof: pd.DataFrame = None
    df_most_deviating_overview: pd.DataFrame = None
    df_most_deviating_overview: pd.DataFrame = None


state = State(corpus_folder='/data/westac')

output_notebook(hide_banner=True)

# %% tags=[] vscode={"end_execution_time": "2020-08-31T18:34:55.995Z", "start_execution_time": "2020-08-31T18:34:55.854Z"}


importlib.reload(vectorize_corpus_gui)
importlib.reload(notebook_utility)

vectorize_corpus_gui.display_gui(state.corpus_folder, '*sparv4.csv.zip', generated_callback=None)

# %% tags=[] vscode={"end_execution_time": "2020-08-31T18:34:55.995Z", "start_execution_time": "2020-08-31T18:34:55.854Z"}
importlib.reload(load_corpus_gui)
importlib.reload(notebook_utility)


def load_succeeded(_v_corpus: VectorizedCorpus, _corpus_tag, output):

    output.clear_output()
    try:
        global state

        state.corpus = _v_corpus
        state.df_gof = gof.compute_goddness_of_fits_to_uniform(state.corpus, 10000, verbose=False)
        state.df_most_deviating_overview = gof.compile_most_deviating_words(state.df_gof, n_count=10000)

        if os.environ.get('VSCODE_LOGS', None) is None:
            _ = (
                notebook_utility.OutputsTabExt(["GoF", "GoF (abs)", "Plots", "Slopes"])
                .display()
                .display_fx_result(0, display_grid, state.df_gof)
                .display_fx_result(
                    1, display_grid, state.df_most_deviating_overview[['l2_norm_token', 'l2_norm', 'abs_l2_norm']]
                )
                .display_fx_result(2, gof.plot_metrics, state.df_gof, plot=False, lazy=True)
                .display_fx_result(
                    3,
                    gof.plot_slopes,
                    state.corpus,
                    state.df_most_deviating_overview,
                    "l2_norm",
                    600,
                    600,
                    plot=False,
                    lazy=True,
                )
            )

    except Exception as ex:
        with output:
            print(ex)


load_corpus_gui.display_gui(state.corpus_folder, load_succeeded)

# %%

xxx_word_trends_gui.display_gui(state=state)

# %%

importlib.reload(word_trend_plot_gui)

most_deviating = gof.get_most_deviating_words(state.df_gof, 'l2_norm', n_count=5000, ascending=False, abs_value=True)

word_trend_plot_gui.display_gui(state.corpus, most_deviating)



import pandas as pd
from ipyaggrid import Grid  # pylint: disable=unused-import
from IPython.display import display

from . import data_compilers

NAME = "Table"

compile = data_compilers.compile_year_token_vector_data  # pylint: disable=redefined-builtin


def setup(container, **_):  # pylint: disable=unused-argument
    pass


def plot(data, **_):  # pylint: disable=unused-argument

    df = pd.DataFrame(data=data)
    df = df[['year'] + [x for x in df.columns if x != 'year']].set_index('year')
    display(df)



import bokeh

from . import data_compilers

NAME = "Bar"

compile = data_compilers.compile_year_token_vector_data  # pylint: disable=redefined-builtin


def setup(container, **kwargs):  # pylint: disable=unused-argument
    pass


def plot(data, **kwargs):
    container = kwargs['container']
    container.data_source.data.update(data)
    bokeh.io.push_notebook(handle=container.handle)
